﻿using System;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Timers;
using System.Windows.Documents;
using System.Windows.Forms;
using System.Windows.Markup;
using HighVoltz.HBRelog.Tasks;
using RichTextBox = System.Windows.Controls.RichTextBox;
using Timer = System.Timers.Timer;

namespace HighVoltz.HBRelog.WebServer
{
    class WebServerRequestHandle : HttpServer
    {
        public WebServerRequestHandle(int port)
            : base(port)
        {
            if (!Directory.Exists("html"))
                Directory.CreateDirectory("html");
            cmdChecker.Interval = 500;
            cmdChecker.Enabled = true;
            cmdChecker.Elapsed += new ElapsedEventHandler(OnCmdCheckerEvent);
        }

        public WebServerRequestHandle(int port, bool refreshTmpPngOnRequest)
            : base(port)
        {
            this.refreshTmpPngOnRequest = refreshTmpPngOnRequest;
            if (!Directory.Exists("html"))
                Directory.CreateDirectory("html");
            cmdChecker.Interval = 500;
            cmdChecker.Enabled = true;
            cmdChecker.Elapsed += new ElapsedEventHandler(OnCmdCheckerEvent);
        }
        
        public bool refreshTmpPngOnRequest = false;

        Timer cmdChecker = new Timer();
        private static void OnCmdCheckerEvent(object source, ElapsedEventArgs e)
        {
            if (File.Exists(ApplicationPath + @"\html\.hbcmd"))
            {
                string cmdString;
                using (StreamReader reader = new StreamReader(ApplicationPath + @"\html\.hbcmd"))
                {
                    cmdString = reader.ReadToEnd();
                    reader.Close();
                }
                File.Delete(ApplicationPath + @"\html\.hbcmd");
                string[] allCmdPairs = cmdString.Split(';');
                foreach (string cmdPairs in allCmdPairs)
                {
                    if (!cmdPairs.Contains('=')) break;
                    string[] cmd = cmdPairs.Split('=');
                    if (cmd[1] == "False")
                        HbRelogManager.Settings.CharacterProfiles[Convert.ToInt32(cmd[0])].Start();
                    if (cmd[1] == "True")
                        HbRelogManager.Settings.CharacterProfiles[Convert.ToInt32(cmd[0])].Stop();

                }
            }
        }

        public override void handleGETRequest(HttpProcessor p)
        {
            finalRequestHandle(p, null);
        }

        public override void handlePOSTRequest(HttpProcessor p, StreamReader inputData)
        {
            finalRequestHandle(p, inputData);

        }

        private void finalRequestHandle(HttpProcessor p, StreamReader inputData)
        {
            string requestedFile = p.http_url;

            string getParam = "";
            string postParam = "";

            //delete / or \ at the beginning
            if (requestedFile.StartsWith("\\") || requestedFile.StartsWith("/"))
                requestedFile = requestedFile.Substring(1);

            if (requestedFile.Contains("?"))
            {
                getParam = requestedFile.Substring(requestedFile.IndexOf("?") + 1);
                requestedFile = requestedFile.Remove(requestedFile.IndexOf("?"));
            }

            if (inputData != null)
            {
                postParam = inputData.ReadToEnd();
            }

            //standart file
            if (requestedFile == "")
                requestedFile = "index.php";

            //set extention
            string requestedFileExtention = requestedFile.Substring(requestedFile.LastIndexOf(".") + 1);


            if (!File.Exists(ApplicationPath + @"\html\" + requestedFile) || requestedFile.StartsWith("."))
            {
                string message = "<h1>404 - File not found</h1><br>\"" +requestedFile + "\" was not found.";

                message = MakeHeader("HTTP/1.1", MIME.getMIME("html"), message.Length, " 404 Not Found") + message;
                p.outputStream.Write(message);
                return;
            }


            //picture?
            if (MIME.isPicture(requestedFileExtention))
                sendPicture(p, requestedFileExtention);

            try
            {
                string outputString = "";
                string cgiHeader = "";
                string file = ApplicationPath + @"\html\" + requestedFile;

                using (StreamReader reader = new StreamReader(file))
                {
                    outputString = reader.ReadToEnd();
                    reader.Close();
                }
                if (requestedFileExtention == "php")
                {
                    outputString = "<?php " +
                                   "parse_str(\"" + getParam + "\", $_GET);" +
                                   "parse_str(\"" + postParam + "\", $_POST);" +
                                   "parse_str(\"" + GenerateHBVars() + "\", $_HB);" +
                                   "parse_str(\"" + GenerateCookieVars(p.httpHeaders) + "\", $_COOKIE);" +
                                   "parse_str(\"" + TaskMonitor() + "\", $_TASKMONITOR);" +
                                   "parse_str(\"" + GetLog() + "\", $_LOG);" +
                                   "?>" + outputString;
                    file = ApplicationPath + @"\html\.tmp." + requestedFileExtention;
                    StreamWriter writer = new StreamWriter(file);
                    writer.Write(outputString);
                    writer.Close();

                    ////PHP
                    ProcessStartInfo processStartInfo = new ProcessStartInfo((ApplicationPath + @"\php\php-cgi.exe"), "\"" + file + "\"");
                    processStartInfo.CreateNoWindow = true;
                    processStartInfo.UseShellExecute = false;
                    processStartInfo.RedirectStandardError = true;
                    processStartInfo.RedirectStandardOutput = true;

                    // execute
                    Process process = Process.Start(processStartInfo);
                    process.WaitForExit();

                    // read outputs
                    string stdOutput = process.StandardOutput.ReadToEnd();
                    string stdError = process.StandardError.ReadToEnd();

                    outputString = stdOutput;
                    if (stdError.Length > 0)
                        outputString = stdError;

                    //Cut of CGI Part
                    int breakPos = outputString.IndexOf("\r\n\r\n");
                    cgiHeader = outputString.Remove(breakPos);
                    outputString = outputString.Substring(breakPos + 4);
                }

                //create header and send!
                outputString = MakeHeader("HTTP/1.1", cgiHeader, outputString.Length, " 200 OK") + outputString;
                p.outputStream.Write(outputString);
            }
            catch (Exception e)
            {
                string message = "An Exception Occurred : " + e.ToString().Remove(e.ToString().IndexOf('\n'));

                message = MakeHeader("HTTP/1.1", MIME.getMIME("html"), message.Length, " 500 Server Error") + message;
                p.outputStream.Write(message);
            }
        }

        private void sendPicture(HttpProcessor p, string requestedFileExtention)
        {
            string path = p.http_url;
            if (path.IndexOf("?") > 0)
                path = path.Remove(path.IndexOf("?"));

            if (path.EndsWith("tmp.png") && refreshTmpPngOnRequest)
                SaveCaptureScreen("html\\tmp.png");

            try
            {
                byte[] picture;

                using (
                    FileStream fstream = new FileStream(ApplicationPath + @"\html\" + path, FileMode.Open,
                                                        FileAccess.Read))
                {
                    //The tempSplitArray //recieves the request from the browser
                    picture = new byte[(long) fstream.Length];

                    fstream.Read(picture, 0, (int) fstream.Length);
                    fstream.Close();
                    ;

                }
                byte[] header = new byte[0];
                if (path.EndsWith(".png"))
                    header = StringToByteArray(MakeHeader("HTTP/1.1", MIME.getMIME(requestedFileExtention), picture.Length, " 200 OK"));

                byte[] finishArray = new byte[picture.Length + header.Length];

                int i = 0;
                foreach (byte b in finishArray)
                {
                    if (i < header.Length)
                        finishArray[i] = header[i];
                    else
                        finishArray[i] = picture[i - header.Length];
                    i++;
                }
                p.outputStream.BaseStream.Write(finishArray, 0, finishArray.Length);
            }
            catch (Exception e)
            {
                string message = "An Exception Occurred : " + e.ToString().Remove(e.ToString().IndexOf('\n'));

                message = MakeHeader("HTTP/1.1", MIME.getMIME(""), message.Length, " 404 Not Found") + message;
                p.outputStream.Write(message);
            }
        }
        
        private string MakeHeader(string sHttpVersion, string sMIMEHeader, int iTotBytes, string sStatusCode)
        {

            String header = "";

            // if Mime type is not provided set default to text/html
            if (sMIMEHeader.Length == 0)
            {
                sMIMEHeader = "text/html";  // Default Mime Type is text/html
            }

            header = header + sHttpVersion + sStatusCode + "\r\n";
            header = header + "Server: cx1193719-b\r\n";
            header = header + "Content-Type: " + sMIMEHeader + "\r\n";
            header = header + "Accept-Ranges: bytes\r\n";
            header = header + "Content-Length: " + iTotBytes + "\r\n\r\n";
            
            return header;
        }

        //HELPER

        private string GenerateCookieVars(Hashtable httpHeaders)
        {
            string result = "";
            foreach (DictionaryEntry entry in httpHeaders)
            {
                if (entry.Key.ToString() == "Cookie")
                {
                    if (entry.Value.ToString().Contains("="))
                    {
                        string[] cookie = entry.Value.ToString().Split(';');
                        foreach (string cookieData in cookie)
                        {
                            string[] values = cookieData.Split('=');
                            result += values[0].Replace(" ", "") + "=" + values[1] + "&";
                        }
                    }
                }
            }
            return result;
        }

        private string GetLog()
        {
            RichTextBox rtb = MainWindow.Instance.LogTextBox;
            var messageTr = new TextRange(rtb.Document.ContentStart, rtb.Document.ContentEnd);
            return "log=" + ReplaceSpecialCharacters(messageTr.Text.Replace("[", "<br>[").Substring(4));
        }

        private string ReplaceSpecialCharacters(string Input)
        {
            Input = Regex.Replace(Input, "&amp;", "&");
            Input = Regex.Replace(Input, "&apos;", "'");
            Input = Regex.Replace(Input, "&#x22;", "\"");
            Input = Regex.Replace(Input, "&#x2039;", "<");
            Input = Regex.Replace(Input, "&#x203A;", ">");
            Input = Regex.Replace(Input, "&gt;", ">=");
            Input = Regex.Replace(Input, "&lt;", "<=");

            return Input;
        }

        private string TaskMonitor()
        {
            int i = 0;
            string result = "";
            string[] monitoredProcessNames = { "Wow", "Honorbuddy", "ARelog", "HBRelog", "WinAuth", "WoWMonitor", "WowError" };
            foreach (string monitoredProcessName in monitoredProcessNames)
            {
                Process[] monitoredProcesses = Process.GetProcessesByName(monitoredProcessName);
                foreach (Process monitoredProcess in monitoredProcesses)
                {
                    string memory = ((monitoredProcess.WorkingSet64 / 1024) / 1024).ToString() + "MB";
                    string name = monitoredProcessName;
                    result += "task[" + i + "][name] = " + name + "&";
                    result += "task[" + i + "][memory] = " + memory + "&";
                    i++;
                }
            }
            result += "task[Count] = " + i;
            return result;
        }

        private string GenerateHBVars()
        {
            string result = "";
            result += "Count=" +  HbRelogManager.Settings.CharacterProfiles.Count().ToString();
            int i = 0;
            foreach (CharacterProfile character in HbRelogManager.Settings.CharacterProfiles)
            {
                result += "&char[" + i + "][isRunning]=" + character.IsRunning.ToString();
                result += "&char[" + i + "][IsPaused]=" + character.IsPaused.ToString();
                result += "&char[" + i + "][Status]=" + character.Status;
                result += "&char[" + i + "][Tooltip]=" + character.Tooltip;
                result += "&char[" + i + "][BotInfoTooltip]=" + character.BotInfoTooltip;

                result += "&char[" + i + "][Settings][ProfileName]=" + character.Settings.ProfileName;

                result += "&char[" + i + "][Settings][HonorbuddySettings][BotBase]=" + character.Settings.HonorbuddySettings.BotBase;
                result += "&char[" + i + "][Settings][HonorbuddySettings][CustomClass]=" + character.Settings.HonorbuddySettings.CustomClass;
                result += "&char[" + i + "][Settings][HonorbuddySettings][HonorbuddyPath]=" + character.Settings.HonorbuddySettings.HonorbuddyPath;
                result += "&char[" + i + "][Settings][HonorbuddySettings][HonorbuddyProfile]=" + character.Settings.HonorbuddySettings.HonorbuddyProfile;

                result += "&char[" + i + "][Settings][WowSettings][AcountName]=" + character.Settings.WowSettings.AcountName;
                result += "&char[" + i + "][Settings][WowSettings][CharacterName]=" + character.Settings.WowSettings.CharacterName;
                result += "&char[" + i + "][Settings][WowSettings][WowPath]=" + character.Settings.WowSettings.WowPath;

                int j = 0;
                foreach (BMTask task in character.Tasks)
                {
                    result += "&char[" + i + "][Tasks][" + j + "][IsDone]=" + task.IsDone;
                    result += "&char[" + i + "][Tasks][" + j + "][IsRunning]=" + task.IsRunning;
                    result += "&char[" + i + "][Tasks][" + j + "][Name]=" + task.Name;
                    result += "&char[" + i + "][Tasks][" + j + "][ToolTip]=" + task.ToolTip;
                    result += "&char[" + i + "][Tasks][" + j + "][Help]=" + task.Help;

                    if (Object.ReferenceEquals(task.GetType(), new ChangeProfileTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=ChangeProfileTask";
                        result += "&char[" + i + "][Tasks][" + j + "][Bot]=" +  ((ChangeProfileTask)task).Bot;
                        result += "&char[" + i + "][Tasks][" + j + "][ProfilePath]=" +  ((ChangeProfileTask)task).ProfilePath;
                    }
                    if (Object.ReferenceEquals(task.GetType(), new IdleTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=IdleTask";
                        result += "&char[" + i + "][Tasks][" + j + "][Minutes]=" + ((IdleTask)task).Minutes;
                        result += "&char[" + i + "][Tasks][" + j + "][RandomMinutes]=" + ((IdleTask)task).RandomMinutes;
                    }
                    if (Object.ReferenceEquals(task.GetType(), new StartProfileTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=StartProfileTask";
                        result += "&char[" + i + "][Tasks][" + j + "][ProfileName]=" + ((StartProfileTask)task).ProfileName;
                    }
                    if (Object.ReferenceEquals(task.GetType(), new StopProfileTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=StopProfileTask";
                        result += "&char[" + i + "][Tasks][" + j + "][ProfileName]=" + ((StopProfileTask)task).ProfileName;
                    }
                    if (Object.ReferenceEquals(task.GetType(), new WaitTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=WaitTask";
                        result += "&char[" + i + "][Tasks][" + j + "][Minutes]=" + ((WaitTask)task).Minutes;
                        result += "&char[" + i + "][Tasks][" + j + "][RandomMinutes]=" + ((WaitTask)task).RandomMinutes;
                    }
                    if (Object.ReferenceEquals(task.GetType(), new LogonTask().GetType()))
                    {
                        result += "&char[" + i + "][Tasks][" + j + "][GetType]=LogonTask";
                        result += "&char[" + i + "][Tasks][" + j + "][BotBase]=" + ((LogonTask)task).BotBase;
                        result += "&char[" + i + "][Tasks][" + j + "][CharacterName]=" + ((LogonTask)task).CharacterName;
                        result += "&char[" + i + "][Tasks][" + j + "][CustomClass]=" + ((LogonTask)task).CustomClass;
                        result += "&char[" + i + "][Tasks][" + j + "][HonorbuddyPath]=" + ((LogonTask)task).HonorbuddyPath;
                        result += "&char[" + i + "][Tasks][" + j + "][ProfilePath]=" + ((LogonTask)task).ProfilePath;
                        result += "&char[" + i + "][Tasks][" + j + "][Server]=" + ((LogonTask)task).Server;
                    }
                    j++;
                }
                i++;
            }
            return result;
        }

        private void SaveCaptureScreen(string path)
        {
            var image = new Bitmap(Screen.PrimaryScreen.Bounds.Width, Screen.PrimaryScreen.Bounds.Height, PixelFormat.Format32bppArgb);
            var gfx = Graphics.FromImage(image);
            gfx.CopyFromScreen(Screen.PrimaryScreen.Bounds.X, Screen.PrimaryScreen.Bounds.Y, 0, 0, Screen.PrimaryScreen.Bounds.Size, CopyPixelOperation.SourceCopy);

            int size = 1200;
            int x, y;

            if (image.Width > image.Height)
            {
                x = size;
                y = Convert.ToInt32(((float)image.Height / (float)image.Width * size));
            }
            else
            {
                y = size;
                x = Convert.ToInt32(((float)image.Width / (float)image.Height * size));
            }
            Image ret = new Bitmap(x, y);
            Graphics g = Graphics.FromImage(ret);
            g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
            g.DrawImage(image, 0, 0, x, y);
            g.Dispose();
            ret.Save(path, ImageFormat.Png);
        }

        private static string ApplicationPath
        {
            get
            {
                return Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
            }
        }

        private byte[] StringToByteArray(string str)
        {
            ASCIIEncoding enc = new ASCIIEncoding();
            return enc.GetBytes(str);
        }
    }
}
